#!/usr/bin/python3
#
#           fistr2rad.py
#
#       fistr形式のmeshをradioss形式に変換する
#
#   23/11/15    新規作成
#      12/06    SPHの取得を追加。
#               デフォルトのmsh, cntファイル名をhcemw_ctrl.datから
#               取得するように修正。
#               optionにversionとunitを追加
#      12/07    getSurfaceGroupDict,createGroupSurfaceLines:
#               SGRPを取得して、rad形式変換を追加
#      12/09    rad形式中に、BCS,INIVEL, INTER/TYPE7, MAT, PROP, PART
#               を追加し、そのままで、計算できる様に修正。
#      12/10    sphをstartに追加。
#      12/12    ENGINEファイル作成を追加
#      12/20    SPHのmp算出をegrp全体質量/粒子数で算出する様に修正
#      12/29    塑性材料（PLASTIC, MISES, MULTILINEAR）の変換を追加
#   24/01/06    単位の取得をcurrDir内から取得に変更。
#               option「-w」を追加。（workFolderを取得）
#

import sys
import os
import datetime
import convertMesh as cm
import geometryFunc as geom
import pyFistr

meshConts = []
radConts = []
engConts = []

#
#  要素Typeとgrp名の対応
elmGrpDict = {
  # rad要素   radGroup名
    "SH3N":   "GRSH3N/SH3N",       #三角形shell 1次
    "SHELL":  "GRSHEL/SHEL",       #三角形shell 1次
    "TETRA4": "GRBRIC/BRIC",       #四面体1次
    "BRICK":  "GRBRIC/BRIC"        #6面体1次
    }

#
#  rad要素名と要素typeの辞書
elmTypeDict = {
    "SH3N":   "shell",       #三角形shell 1次
    "SHELL":  "shell",       #三角形shell 1次
    "TETRA4": "solid",       #四面体1次
    "BRICK":  "solid",       #6面体1次
    "SPHCEL": "sph"          #粒子
    }

#
#  node順入れ替え辞書
fistr2rad_elDict = {
    731: ["SH3N",   [1,2,3], cm.order_3],
    761: ["SH3N",   [1,2,3], cm.order_3],
    741: ["SHELL",  [1,2,3,4], cm.order_4],
    781: ["SHELL",  [1,2,3,4], cm.order_4],
    341: ["TETRA4", [1,2,4,3], cm.order_4],
    361: ["BRICK",  [1,2,3,4,5,6,7,8], cm.order_8]
    }

#
#  unitNoの辞書
unitDict = {
    #unitNo   unit
    "1": ["kg", "m", "s"],      #MKS
    "2": ["ton", "mm", "s"],
    "3": ["kg", "mm", "ms"],
    "4": ["g", "cm", "s"]       #CGS
    }

#
#  checkStrLength10
def checkStrLength10(vals):
    """ 値が20文字以内かチェック"""
    newVals = []
    for val in vals:
        if len(val) >= 10:
            num = float(val)
            newVal = "%10.3e" % num
            newVals.append(newVal)
        else:
            newVals.append(val)
    return newVals

#
#  checkStrLength20
def checkStrLength20(vals):
    """ 値が20文字以内かチェック"""
    newVals = []
    for val in vals:
        if len(val) >= 20:
            num = float(val)
            newVal = "%20.10e" % num
            newVals.append(newVal)
        else:
            newVals.append(val)
    return newVals

#
#  getCntMshFileName
def getCntMshFileName(workDir):
    """ hecmw_ctrl.datファイルからcnt, mshファイル名を取得する"""
    """ hecmw_ctrl.datから、cnt, mshファイル名を取得して返す"""
    cntName = "FistrModel.cnt"
    mshName = "FistrModel.msh"
    fileName = workDir + os.sep + "hecmw_ctrl.dat"
    if os.path.exists(fileName) == False:
        return cntName, mshName
    f=open(fileName, encoding="utf-8"); lines=f.readlines(); f.close()
    #cntFile
    i = pyFistr.getNextHeaderName(lines, 0, "!CONTROL", "NAME=fstrCNT")
    i += 1
    if i < len(lines):
        cntName = lines[i].split()[0]
    #mshFile
    i = pyFistr.getNextHeaderName(lines, 0, "!MESH", "TYPE=HECMW-ENTIRE")
    i += 1
    if i < len(lines):
        mshName = lines[i].split()[0]
    return cntName, mshName

#
#  getHeaderDataOfCard
#----------------------
def getHeaderDataOfCard(headerData, cardName):
    """ headerData内から、全てのcardNameのdataを取得する"""
    cardData = []
    for header, data in headerData:
        words = pyFistr.deleteSp(header).split(",")
        if words[0] == cardName:
            cardData.append([header, data])
    return cardData

#
#  getNodeDataDict
#--------------
def getNodeDataDict(headerData):
    """ nodeDataを取得する。
    ngrp名に「dummy」の文字があるnodeDataは取得しない。
    nodeDict = {<nodeNo>:[x,y,z]}"""
    nodeDict = {}
    for header, data in headerData:
        words = pyFistr.deleteSp(header).split(",")
        if words[0].upper() == "!NODE":
            for vals in data:
                nodeDict[vals[0]] = vals[1:]
    return nodeDict

#
#  getElementDataDict
#---------------------
def getElementDataDict(headerData):
    """ 要素dataを取得する
    elmDict = {<elmNo>:[[<n1, n2,...], <elmType>]}"""
    elmDict = {}
    for header, data in headerData:
        words = pyFistr.deleteSp(header).split(",")
        if words[0].upper() == "!ELEMENT":
            elmType = int(pyFistr.getValName(header, "TYPE"))
            for vals in data:
                elmNo = vals[0]
                nodeNos = vals[1:]
                elmDict[elmNo] = [nodeNos, elmType]
    return elmDict

#
#  getNodeElementDict
#----------------------
def getNodeElementDict(elmDict):
    """ nodeが属する要素Noを取得する
    nodeElmDict = {<nodeNo>:[elmNo_1, elmNo_2,..], ..}"""
    nodeElmDict = {}
    for elmNo in elmDict.keys():
        nodeNos = elmDict[elmNo][0]
        for nodeNo in nodeNos:
            if not nodeNo in nodeElmDict.keys():
                nodeElmDict[nodeNo] = set([elmNo])
            else:
                nodeElmDict[nodeNo].add(elmNo)
    return nodeElmDict

#
#  deleteDummyNodesChangeElmType
#--------------------------------
def deleteDummyNodesChangeElmType(nodeDict, elmDict, ngrpDict):
    """ dummyNodeを持っている要素（761、781）のdummyNodeを削除し、
    要素typeを731、741に変更して返す。"""
    dummyNodes = []
    for elmNo in elmDict.keys():
        nodes = elmDict[elmNo][0]
        elmType = elmDict[elmNo][1]
        if elmType == 761:
            dummyNodes += nodes[3:]
            elmDict[elmNo] = [nodes[:3], 731]
        elif elmType == 781:
            dummyNodes += nodes[4:]
            elmDict[elmNo] = [nodes[:4], 741]
    #dummyNodeを削除
    dummyNodes = set(dummyNodes)
    for nodeNo in dummyNodes:
        a = nodeDict.pop(nodeNo)
    #ngrpDicからdummyNodeを削除
    for ngrpName in ngrpDict.keys():
        nodeNos = ngrpDict[ngrpName]
        newNodes = set(nodeNos) - set(dummyNodes)
        ngrpDict[ngrpName] = list(newNodes)
    #ngrpDict内のnode数=0のgrpは削除する
    names = list(ngrpDict.keys())
    for name in names:
        if len(ngrpDict[name]) == 0:
            a = ngrpDict.pop(name)
    return nodeDict, elmDict, ngrpDict

#
#  getShellThicknessDict
#------------------------
def getShellThicknessDict(headerData):
    """ shell要素がある場合は、板圧を取得する
    shellThDict = {<egrp>: [<板圧>, <積分点>], ...}"""
    shellThDict = {}
    for header, data in headerData:
        words = pyFistr.deleteSp(header).split(",")
        if words[0] == "!SECTION":
            elmType = pyFistr.getValName(header, "TYPE")
            if elmType == "SHELL":
                egrp = pyFistr.getValName(header, "EGRP")
                ws = pyFistr.deleteSp(data[0]).split(",")
                th = ws[0]      #板圧
                N = ws[1]       #積分点
                shellThDict[egrp] = [th, N]
    return shellThDict

#
#  getContactPair
#-----------------
def getContactPair(headerData):
    """ contactPairを取得する"""
    contactDict = {}
    for header, data in headerData:
        words = pyFistr.deleteSp(header).split(",")
        if words[0] == "!CONTACT PAIR" or words[0] == "!CONTACTPAIR":
            contactName = pyFistr.getValName(header, "NAME")
            pairNames = pyFistr.deleteSp(data[0]).split(",")
            contactDict[contactName] = pairNames
    return contactDict

#
#  getNodeGroupDict
#-------------------
def getNodeGroupDict(headerData):
    """ nodeGroupを取得する。
    group名に「dummy」の文字が入っている場合を除く。
    ngrpDict = {<ngrp名>:{n1, n2,...}}"""
    ngrpDict = {}
    for header, data in headerData:
        words = pyFistr.deleteSp(header).split(",")
        if words[0] == "!NODE":
            name = pyFistr.getValName(header, "NGRP")
            if name != "":
                if not name in ngrpDict.keys():
                    ngrpDict[name] = set([])
                for vals in data:
                    nodeNo = vals[0]
                    ngrpDict[name].add(nodeNo)
        elif words[0] == "!NGROUP":
            name = pyFistr.getValName(header, "NGRP")
            if not name in ngrpDict.keys():
                ngrpDict[name] = set([])
            nodeSet = set(data)
            ngrpDict[name] = ngrpDict[name] | nodeSet
    return ngrpDict

#
#  getSurfaceGroupDict
#----------------------
def getSurfaceGroupDict(headerData, elmDict):
    """ surfaceGroupを取得する
    surfDict = {<SGRP名>:[<nodes0>, <nodes1>,...]}"""
    surfDict = {}
    for header, data in headerData:
        words = pyFistr.deleteSp(header).split(",")
        if words[0] == "!SGROUP":
            name = pyFistr.getValName(header, "SGRP")
            if name != "":
                if not name in surfDict.keys():
                    surfDict[name] = []
                for i in range(0, len(data), 2):
                    elmNo = data[i]
                    faceNo = data[i+1]
                    elNodes = elmDict[elmNo][0]
                    elType = str(elmDict[elmNo][1])
                    nodes = cm.fistr_elFaceNode(elType, elNodes, faceNo)
                    surfDict[name] += [nodes]
    return surfDict

#
#  getElementGroupDict
#----------------------
def getElementGroupDict(headerData):
    """ egrpを取得する
    egrpDict = {<egrp名>:{e1, e2,...}}"""
    egrpDict = {}
    for header, data in headerData:
        words = pyFistr.deleteSp(header).split(",")
        if words[0] == "!ELEMENT":
            name = pyFistr.getValName(header, "EGRP")
            if name != "":
                if not name in egrpDict.keys():
                    egrpDict[name] = set([])
                elms = map(lambda x: x[0], data)
                egrpDict[name] = egrpDict[name] | set(elms)
        if words[0] == "!EGROUP":
            name = pyFistr.getValName(header, "EGRP")
            if not name in egrpDict.keys():
                egrpDict[name] = set([])
            elmsSet = set(data)
            egrpDict[name] = egrpDict[name] | elmsSet
    return egrpDict

#
#  renumberNode
#---------------
def renumberNode(nodeDict, elmDict, ngrpDict):
    """ nodeNoをrenumberingする"""
    nodeNos = list(nodeDict.keys())
    nodeNos.sort()
    newNodeNoDict = {}
    newNodeNo = 1
    for nodeNo in nodeNos:
        newNodeNoDict[nodeNo] = newNodeNo
        newNodeNo += 1
    newNodeDict = {}
    for nodeNo in nodeDict.keys():
        newNo = newNodeNoDict[nodeNo]
        newNodeDict[newNo] = nodeDict[nodeNo]
    for elmNo in elmDict.keys():
        nodes = elmDict[elmNo][0]
        newNodes = []
        for nodeNo in nodes:
            newNo = newNodeNoDict[nodeNo]
            newNodes.append(newNo)
        elmDict[elmNo][0] = newNodes
    for ngrpName in ngrpDict.keys():
        nodes = ngrpDict[ngrpName]
        newNodes = []
        for nodeNo in nodes:
            newNo = newNodeNoDict[nodeNo]
            newNodes.append(newNo)
        ngrpDict[ngrpName] = newNodes
    return newNodeDict, elmDict, ngrpDict

#
#  convertElement
#-----------------
def convertElement(elmDict):
    """ 要素の並び順をrad形式に変換する
    radDict = {<elmNo>:[[n1, n2,...], <elmType>]}"""
    radDict = {}
    for elmNo in elmDict.keys():
        nodeNos = elmDict[elmNo][0]
        elmType = elmDict[elmNo][1]
        (radType, n, func) = fistr2rad_elDict[elmType]
        newNos = func(nodeNos, n)
        radDict[elmNo] = [newNos, radType]
    return radDict

#
#  getEgrpMaterialDictFromCnt
#-----------------------------
def getEgrpMaterialDictFromCnt(cntHeaderData, headerData):
    """ egrpと材料定数をcntFileから取得
    matDict = {<egrp>:[<matName>, E, nu, rho, alpha, []],...}"""
    matValDict = {}
    for header, data in cntHeaderData:
        words = pyFistr.deleteSp(header).split(",")
        if words[0] == "!MATERIAL":
            matName = pyFistr.getValName(header, "NAME")
            #                      E   nu  rho alpha []
            matValDict[matName] = ["", "", "", "", []]
        elif words[0] == "!ELASTIC":
            ws = pyFistr.deleteSp(data[0]).split(",")
            matValDict[matName][0] = ws[0]
            matValDict[matName][1] = ws[1]
        elif words[0] == "!DENSITY":
            ws = pyFistr.deleteSp(data[0]).split(",")
            matValDict[matName][2] = ws[0]
        elif words[0] == "!EXPANSION_COEFF":
            ws = pyFistr.deleteSp(data[0]).split(",")
            matValDict[matName][3] = ws[0]
        elif words[0] == "!PLASTIC":
            matValDict[matName][4] = [header, data]
    #egrpを取得
    matDict = {}
    for header, data in headerData:
        words = pyFistr.deleteSp(header).split(",")
        if words[0] == "!SECTION":
            egrp = pyFistr.getValName(header, "EGRP")
            matName = pyFistr.getValName(header, "MATERIAL")
            if matName in matValDict.keys():
                matDict[egrp] = [matName] + matValDict[matName]
    return matDict

#
#  getEgrpMaterialDict
#------------------------
def getEgrpMaterialDict(headerData):
    """ egrpと材料定数をcntFileから取得
    matDict = {<egrp>:[<matname>, <youngs>, <poisson>, <density>],...}"""
    matDict = {}
    matValDict = {}
    for header, data in headerData:
        words = pyFistr.deleteSp(header).split(",")
        if words[0] == "!SECTION":
            egrp = pyFistr.getValName(header, "EGRP")
            matName = pyFistr.getValName(header, "MATERIAL")
            matDict[egrp] = [matName]
        elif words[0] == "!MATERIAL":
            matName = pyFistr.getValName(header, "NAME")
        elif words[0] == "!ITEM=1":
            matValDict[matName] = pyFistr.deleteSp(data[0]).split(",")
        elif words[0] == "!ITEM=2":
            matValDict[matName] += pyFistr.deleteSp(data[0]).split(",")
    for egrp in matDict.keys():
        matName = matDict[egrp][0]
        if matName in matValDict.keys():
            matVals = matValDict[matName]
            matDict[egrp] += matVals
        else:
            matVals = ["", "", ""]
            matDict[egrp] += matVals
    return matDict

#
#  getGroupIdFromMeshConts
def getGroupIdFromMeshConts(grpName, grp):
    """ meshContsからgrpとgrpNameに対応するgrpIDを取得する"""
    global meshConts
    for line in meshConts:
        words = line.split()
        ws = words[1].split("/")
        grpID = ""
        if len(ws) >= 2:
            #grp(GRNOD, SURF, GRBRIC, etc)とgrpNameが同じ？
            if ws[1] == grp and words[-1] == grpName:
                grpID = ws[-1]
                break
    return grpID


#
#  createRadiossIncMeshFile
#----------------------------
def createRadiossIncMeshFile(fistrDicts, elmRadDict, sphEgrp, incFile):
    """ RadiossIncMeshファイルを作成する"""
    global meshConts
    [nodeDict, elmDict, ngrpDict, surfDict, egrpDict,
     matDict, shellThDict, contactDict, nodeElmDict] = fistrDicts 
    meshConts = []
    lines = []
    lines += createNodeLines(nodeDict)
    lines += createElementLines(elmRadDict, egrpDict, sphEgrp)
    lines += createGroupNodeLines(ngrpDict)
    lines += createGroupSurfaceLines(surfDict)
    lines += createGroupElementLines(elmRadDict, egrpDict, sphEgrp)
    headerLines = createHeaderLines()
    footerLines = createFooterLines()
    lines = headerLines + lines + footerLines
    #file出力
    f = open(incFile, "w", encoding="utf-8"); f.writelines(lines); f.close()

#
#  createNodeLines
#-------------------
def createNodeLines(nodeDict):
    """ node部のlinesを作成する"""
    global meshConts
    lines = []
    lines += ["#--------.---------.---------.---------.---------.---------.---------.---------.---------.---------.\n"]
    lines += ["#  NODES\n"]
    lines += ["#---1----|----2----|----3----|----4----|----5----|----6----|----7----|----8----|----9----|---10----|\n"]
    lines += ["/NODE\n"]
    nodeNos = list(nodeDict.keys())
    nodeNos.sort()
    meshConts += ["#   /NODE   Nodes  " + str(len(nodeNos)) + " nodes\n"]
    for nodeNo in nodeNos:
        line = "%10d" % nodeNo
        x, y, z = nodeDict[nodeNo]
        x, y, z = checkStrLength20([str(x), str(y), str(z)])
        strXYZ = ("%20s" % x) + ("%20s" % y) + ("%20s" % z)
        line += strXYZ + "\n"
        lines.append(line)
    return lines

#
#  createElementLines
#---------------------
def createElementLines(elmDict, egrpDict, sphEgrp):
    """ elementのlinesを作成する"""
    global meshConts
    radDict = {}
    partNo = 1001
    egrpNames = list(egrpDict.keys())
    egrpNames.sort()
    maxElmNo = 0
    for egrpName in egrpNames:
        elmNos = egrpDict[egrpName]
        for elmNo in elmNos:
            nodeNos = elmDict[elmNo][0]
            elmType = elmDict[elmNo][1]
            key = (partNo, elmType, egrpName)
            if not key in radDict.keys():
                radDict[key] = []
            radDict[key].append([elmNo] + nodeNos)
        partNo += 1
    lines = []
    for key in radDict.keys():
        partNo, elmType, egrpName = key
        if egrpName != sphEgrp:
            #通常の要素を作成
            elms = radDict[key] 
            meshConts += ["#   /" + elmType + "/" + str(partNo) + "  Elements  " + str(len(elms)) + " elms  " + egrpName + "\n"]
            lines += ["#--------.---------.---------.---------.---------.---------.---------.---------.---------.---------.\n"]
            lines += ["#  ELEMENT: partID=" + str(partNo) + "  " + egrpName + "\n"]
            lines += ["#---1----|----2----|----3----|----4----|----5----|----6----|----7----|----8----|----9----|---10----|\n"]
            lines += ["/" + elmType + "/" + str(partNo) + "\n"]
            for i in range(len(elms)):
                line = ""
                elmNo = elms[i][0]
                nodeNos = elms[i][1:]
                line += "%10d" % elmNo
                for nodeNo in nodeNos:
                    word = "%10d" % nodeNo
                    line += word
                line += "\n"
                lines += [line]
        else:
            #SPHを作成
            nodeNosSet = set([])
            elms = radDict[key]
            for i in range(len(elms)):
                nodeNos = elms[i][1:]
                nodeNosSet.update(nodeNos)
            nodeNos = list(nodeNosSet)
            nodeNos.sort()
            meshConts += ["#   /SPHCEL/" + str(partNo) + "  Elements  " + str(len(nodeNos)) + " points   " + egrpName + "\n"]
            lines += ["#--------.---------.---------.---------.---------.---------.---------.---------.---------.---------.\n"]
            lines += ["#  SPH: partID=" + str(partNo) + "  " + egrpName + "\n"]
            lines += ["#---1----|----2----|----3----|----4----|----5----|----6----|----7----|----8----|----9----|---10----|\n"]
            lines += ["/SPHCEL/" + str(partNo) + "\n"]
            for i in range(len(nodeNos)):
                lines += [("%10d" % nodeNos[i]) + "\n"]
    return lines

#
#  createGroupNodeLines
#------------------------
def createGroupNodeLines(ngrpDict):
    """ groupNode部を作成"""
    global meshConts
    ngrpNo = 1
    lines = []
    ngrpNames = list(ngrpDict.keys())
    ngrpNames.sort()
    for ngrpName in ngrpNames:
        lines += ["#--------.---------.---------.---------.---------.---------.---------.---------.---------.---------.\n"]
        lines += ["#  groupNode: grNodeID=" + str(ngrpNo) + "  " + ngrpName + "\n"]
        lines += ["#---1----|----2----|----3----|----4----|----5----|----6----|----7----|----8----|----9----|---10----|\n"]
        lines += ["/GRNOD/NODE/" + str(ngrpNo) + "\n"]
        lines += ["grNode: " + ngrpName + "\n"]
        nodeNos = list(ngrpDict[ngrpName])
        meshConts += ["#   /GRNOD/NODE/" + str(ngrpNo) + "  NodeGroup  " + str(len(nodeNos)) + " nodes  " + ngrpName + "\n"]
        for i in range(0, len(nodeNos), 10):
            nodes = nodeNos[i:i+10]
            line = ""
            for nodeNo in nodes:
                word = "%10d" % nodeNo
                line += word
            line += "\n"
            lines += [line]
        ngrpNo += 1
    return lines

#
#  createGroupSurfaceLines
def createGroupSurfaceLines(surfDict):
    """ groupSurfaceを作成する"""
    global meshConts
    sgrpNo = 1
    segID = 1
    lines = []
    sgrpNames = list(surfDict.keys())
    sgrpNames.sort()
    for sgrpName in sgrpNames:
        lines += ["#--------.---------.---------.---------.---------.---------.---------.---------.---------.---------.\n"]
        lines += ["#  groupSurface: grSurfID=" + str(sgrpNo) + "  " + sgrpName + "\n"]
        lines += ["#---1----|----2----|----3----|----4----|----5----|----6----|----7----|----8----|----9----|---10----|\n"]
        lines += ["/SURF/SEG/" + str(sgrpNo) + "\n"]
        lines += ["grSurface: " + sgrpName + "\n"]
        segNodes = surfDict[sgrpName]
        meshConts += ["#   /SURF/SEG/" + str(sgrpNo) + "  SurfaceGroup  " + str(len(segNodes)) + " faces  " + sgrpName + "\n"]
        for nodes in segNodes:
            words = []
            for node in nodes:
                words.append(("%10d" % node))
            lines += [("%10d" % segID) + "".join(words) + "\n"]
            segID += 1
        sgrpNo += 1
    return lines

#
#  createGroupElementLines
#--------------------------
def createGroupElementLines(elmDict, egrpDict, sphEgrp):
    """ groupElement部を作成"""
    global meshConts
    egrpNo = 1
    radDict = {}
    egrpNames = list(egrpDict.keys())
    for egrpName in egrpNames:
        elmNos = egrpDict[egrpName]
        for elmNo in elmNos:
            elmType = elmDict[elmNo][1]     #rad要素type
            grpName = elmGrpDict[elmType]   #radGrp名
            key = (egrpNo, grpName, egrpName)
            if not key in radDict.keys():
                radDict[key] = []
            radDict[key].append(elmNo)
        egrpNo += 1
    lines = []
    for key in radDict.keys():
        egrpNo, grpName, egrpName = key
        if egrpName != sphEgrp:
            lines += ["#--------.---------.---------.---------.---------.---------.---------.---------.---------.---------.\n"]
            lines += ["#  groupElms: grElmID=" + str(egrpNo) + "  " + egrpName + "\n"]
            lines += ["#---1----|----2----|----3----|----4----|----5----|----6----|----7----|----8----|----9----|---10----|\n"]
            lines += ["/" + grpName + "/" + str(egrpNo) + "\n"]
            lines += ["grElm: " + egrpName + "\n"]
            elmNos = radDict[key]
            meshConts += ["#   /" + grpName + "/" + str(egrpNo) + "  ElementGroup  " + str(len(elmNos)) + " elms  " + egrpName + "\n"]
            for i in range(0, len(elmNos), 10):
                line = ""
                elms = elmNos[i:i+10]
                for elmNo in elms:
                    word = "%10d" % elmNo
                    line += word
                line += "\n"
                lines += [line]
    return lines
         
#
#  createHeaderLines
#--------------------
def createHeaderLines():
    global meshConts
    addLines = []
    addLines += ["#--------.---------.---------.---------.---------.---------.---------.---------.---------.---------.\n"]
    addLines += ["#  convert mesh from Fistr to Radioss format\n"]
    addLines += ["#\n"]
    addLines += ["#     mesh contents\n"]
    addLines += meshConts
    addLines += ["#--------.---------.---------.---------.---------.---------.---------.---------.---------.---------.\n"]
    addLines += ["/BEGINE\n"]
    addLines += ["convert mesh\n"]
    return addLines

#
#  createFooterLines
#--------------------
def createFooterLines():
    """ footerを作成"""
    addLines = ["/END\n"]
    addLines += ["#--------.---------.---------.---------.---------.---------.---------.---------.---------.---------.\n"]
    return addLines

#
#  getFixBoundary
#-----------------
def getFixBoundary(cntHeaderData):
    """ boundary(0,0,0)を取得する"""
    global meshConts

    def convStr(fix):
        ans = str(fix[0]) + str(fix[1]) + str(fix[2])
        return ans

    fixNameDict = {}
    fix = [0,0,0]
    name = ""
    for header, data in cntHeaderData:
        words = pyFistr.deleteSp(header).split(",")
        if words[0] == "!BOUNDARY":
            for line in data:
                ws = pyFistr.deleteSp(line).split(",")
                if name != ws[0]:
                    if fix != [0,0,0] and name.find("dummy") < 0:
                        fixNameDict[name] = [convStr(fix)]
                    name = ws[0]
                    fix = [0,0,0]
                if float(ws[3]) == 0.0:
                    col = int(ws[1]) - 1
                    fix[col] = 1
    if fix != [0,0,0] and name.find("dummy") < 0:
        fixNameDict[name] = [convStr(fix)]
    #nodeGrpIDを取得
    for name in fixNameDict.keys():
        grpID = getGroupIdFromMeshConts(name, "GRNOD")
        fixNameDict[name].append(grpID)
    return fixNameDict

#
#  getInitialVelocity
#---------------------
def getInitialVelocity(cntHeaderData):
    """ 初期速度を取得"""
    global meshConts

    def convStr(iniVel):
        ans = list(map(str, iniVel))
        return ans

    iniVelDict = {}
    iniVel = [0,0,0]
    name = ""
    for header, data in cntHeaderData:
        words = pyFistr.deleteSp(header).split(",")
        if words[0] == "!VELOCITY":
            for line in data:
                ws = pyFistr.deleteSp(line).split(",")
                if name != ws[0]:
                    if iniVel != [0,0,0] and name.find("dummy") < 0:
                        iniVelDict[name] = [convStr(iniVel)]
                    name = ws[0]
                    iniVel = [0,0,0]
                col = int(ws[1]) - 1
                iniVel[col] = float(ws[3])
    if iniVel != [0,0,0] and name.find("dummy") < 0:
         iniVelDict[name] = [convStr(iniVel)]
    #nodeGrpIDを取得
    for name in iniVelDict.keys():
        grpID = getGroupIdFromMeshConts(name, "GRNOD")
        iniVelDict[name].append(grpID)
    return iniVelDict

#
#  createRadiossStarterFile
#---------------------------
def createRadiossStarterFile(fistrFiles, radiossFiles, fistrDicts):
    """ radioss starterファイルを作成"""
    global radConts
    radConts = []
    [meshFile, cntFile, sphEgrp, fistrUnit] = fistrFiles
    [incFile, radFile, yearVersion, unit] = radiossFiles
    [nodeDict, elmDict, ngrpDict, surfDict, egrpDict, 
     matDict, shellThDict, contactDict, nodeElmDict] = fistrDicts
    #fistrのcntFileを取得
    f = open(cntFile, encoding="utf-8"); lines = f.readlines(); f.close()
    cntHeaderData = pyFistr.getHeaderData(lines)
    #BOUNDARY(0,0,0)を取得
    fixNameDict = getFixBoundary(cntHeaderData)
    #iniVelを取得
    iniVelDict = getInitialVelocity(cntHeaderData)
    #radFileを作成
    lines  = createRadHeaderLines(cntFile, meshFile)
    lines += createBeginUnitsLines(yearVersion, unit, radFile)
    lines += createMaterialLines(matDict)
    lines += createIncludeMeshLines(incFile)
    lines += createPropertyLines(fistrDicts)
    lines += createPartLines()
    lines += createFixBoundaryLines(fixNameDict)
    lines += createInitialVelocityLines(iniVelDict)
    lines += createContactLines(cntHeaderData, fistrDicts)
    lines += createRadFooter()
    #fileに出力
    f = open(radFile, "w", encoding="utf-8"); f.writelines(lines); f.close()

#
#  createRadHeaderLines
#-----------------------
def createRadHeaderLines(cntFile, meshFile):
    """ radFileのheaderを作成"""
    now = datetime.datetime.now()
    dateList = [str(now.year), str(now.month), str(now.day)]
    addLines  = ["#RADIOSS STARTER\n"]
    addLines += ["#===================================================================================================\n"]
    addLines += ["#  generated by fistr2rad.py\n"]
    addLines += ["#           from " + meshFile + " and " + cntFile + "\n"]
    addLines += ["#                                                " + "/".join(dateList) + "\n"]
    addLines += ["#---1----|----2----|----3----|----4----|----5----|----6----|----7----|----8----|----9----|---10----|\n"]
    return addLines

#
#  createBeginUnitsLines
#------------------------
def createBeginUnitsLines(yearVersion, unit, radFile):
    global radConts
    runName = "_".join(radFile.split("_")[:-1])
    addLines  = ["/BEGIN\n"]
    addLines += [runName + "\n"]
    addLines += [("%10s" % yearVersion) + "\n"]
    addLines += [("%20s" % unit[0]) + ("%20s" % unit[1]) + ("%20s" % unit[2]) + "\n"]
    addLines += [("%20s" % unit[0]) + ("%20s" % unit[1]) + ("%20s" % unit[2]) + "\n"]
    radConts.append("/BEGIN  " + yearVersion + "  unit: " + ",".join(unit) + "\n")
    return addLines

#
#  createMaterialLines
#-----------------------
def createMaterialLines(matDict):
    """ 材料物性値を作成"""
    global radConts
    egrpNames = list(matDict.keys())
    if len(egrpNames) == 0:
        return []
    addLines = []
    addLines += ["#===================================================================================================\n"]
    addLines += ["#  MATERIAL define\n"]
    egrpNames.sort()
    matNo = 1
    matFuncNo = 1
    for egrp in egrpNames:
        matName, E, nu, rho, alpha, plastics = matDict[egrp]
        matYield, matHarden = getPlasticYieldHarden(plastics)
        if matYield == "":
            #弾性材料を作成
            addLines += createElasticMaterialLines(matNo, egrp, matDict)
        elif matYield == "MISES" and matHarden == "MULTILINEAR":
            #mises, multilinearの行を作成
            addLines += createPlasticMisesMultilinearLines(matNo, egrp, matDict, matFuncNo)
            addLines += createMultilinearDataLines(egrp, matDict, matFuncNo)
            matFuncNo += 1
        matNo += 1
    return addLines

#
#  getPlasticYieldHarden
def getPlasticYieldHarden(plastics):
    """ !PLASTICの内容からyield、hardenを取得して返す"""
    if len(plastics) == 0:
        return "", ""
    header = plastics[0]
    matYield = pyFistr.getValName(header, "YIELD")
    matHarden = pyFistr.getValName(header, "HARDEN")
    return matYield, matHarden

#
#  createElasticMaterialLines
def createElasticMaterialLines(matNo, egrp, matDict):
    """ 弾性材料の設定行を作成する"""
    global radConts
    matName, E, nu, rho, alpha, plastics = matDict[egrp]
    E, nu, rho = checkStrLength20([E, nu, rho])
    addLines = []
    addLines += ["#---1----|----2----|----3----|----4----|----5----|----6----|----7----|----8----|----9----|---10----|\n"]
    addLines += ["#  MAT-" + str(matNo) + "  ElementGroup " + egrp + "\n"]
    addLines += ["#--------.---------.---------.---------.\n"]
    addLines += ["/MAT/PLAS_JOHNS/" + str(matNo) + "\n"]
    addLines += [matName + " of " + egrp + "\n"]
    addLines += ["#                rho\n"]
    addLines += [("%20s" % rho) + "\n"]
    addLines += ["#                  E                  nu     Iflag\n"]
    addLines += [("%20s" % E) + ("%20s" % nu) + ("%10d" % 1) + "\n"]
    addLines += ["#            sigma_y                 UTF              epsUTF           eps_p_max            sigmaMax\n"]
    addLines += [("%20s" % "1e30") + ("%20s" % "1e30") + "\n"]
    radConts.append("/MAT/PLAS_JOHNS/" + str(matNo) + "  ElementGroup " + egrp + "\n")
    return addLines

#
#  createPlasticMisesMultilinearLines
def createPlasticMisesMultilinearLines(matNo, egrp, matDict, matFuncNo):
    """ 塑性材料（mises, multiliear)の行を作成"""
    global radConts
    matName, E, nu, rho, alpha, plastics = matDict[egrp]
    E, nu, rho = checkStrLength20([E, nu, rho])
    addLines = []
    addLines += ["#---1----|----2----|----3----|----4----|----5----|----6----|----7----|----8----|----9----|---10----|\n"]
    addLines += ["#  MAT-" + str(matNo) + "  ElementGroup " + egrp + "\n"]
    addLines += ["#--------.---------.---------.---------.\n"]
    addLines += ["/MAT/PLAS_TAB/" + str(matNo) + "\n"]
    addLines += [matName + " of " + egrp + "\n"]
    addLines += ["#                rho\n"]
    addLines += [("%20s" % rho) + "\n"]
    addLines += ["#                  E                  nu           eps_p_max               eps_t               eps_m\n"]
    addLines += [("%20s" % E) + ("%20s" % nu) + "\n"]
    addLines += ["#    nFunc   Fsmooth              C_hard                Fcut               eps_f                  VP\n"]
    addLines += ["         1         0                   0\n"]
    addLines += ["#    Ipfun              Fscale\n"]
    addLines += ["         0                 1.0\n"]
    addLines += ["#  FunctID\n"]
    addLines += [("%10d" % matFuncNo) + "\n"]   #関数No定義
    addLines += ["#           FscaleID\n"]
    addLines += ["                 1.0\n"]
    addLines += ["#          eps_rates\n"]
    addLines += ["                 0.0\n"]
    radConts.append("/MAT/PLAS_TAB/" + str(matNo) + "  ElementGroup " + egrp + "\n")
    return addLines

#
#  createMultilinearDataLines
def createMultilinearDataLines(egrp, matDict, matFuncNo):
    """ multilinearFunctionDataを作成"""
    global radConts
    matName, E, nu, rho, alpha, plastics = matDict[egrp]
    data = plastics[1]
    addLines = []
    addLines += ["#---1----|----2----|----3----|----4----|----5----|----6----|----7----|----8----|----9----|---10----|\n"]
    addLines += ["#  MAT_FUNC-" + str(matFuncNo) + "  ElementGroup " + egrp + "\n"]
    addLines += ["#--------.---------.---------.---------.\n"]
    addLines += ["/FUNCT/" + str(matFuncNo) + "\n"]
    addLines += ["#   plastic strain:x      yield stress:y\n"]
    for line in data:
        words = pyFistr.deleteSp(line).split(",")
        stress, strain = words[:2]
        addLines += [("%20s" % strain) + ("%20s" % stress) + "\n"]
    radConts.append("/FUNCT/" + str(matFuncNo) + "  ElementGroup " + egrp + "\n")
    return addLines

#
#  createIncludeMeshLines
#-------------------------
def createIncludeMeshLines(incFile):
    """ include行を作成"""
    global meshConts, radConts
    addLines = []
    addLines += ["#===================================================================================================\n"]
    addLines += ["#  convert mesh from Fistr to Radioss format\n"]
    addLines += ["#\n"]
    addLines += ["#     mesh contents\n"]
    addLines += meshConts
    addLines += ["#--------.---------.---------.---------.---------.---------.---------.---------.---------.---------.\n"]
    addLines += ["#include " + incFile + "\n"]
    addLines += ["#--------.---------.---------.---------.---------.---------.---------.---------.---------.---------.\n"]
    addLines += ["#\n"]
    radConts.append("#include " + incFile + "\n")
    return addLines

#
#  createPropertyLines
#----------------------
def createPropertyLines(fistrDicts):
    """ 要素のpropertyを作成する"""
    global meshConts, radConts
    [nodeDict, elmDict, ngrpDict, surfDict, egrpDict, 
     matDict, shellThDict, contactDict, nodeElmDict] = fistrDicts
    #要素Typeを取得
    egrpElmDict = {}
    for line in meshConts:
        words = line.split()
        if words[2] == "Elements":
            elmType = words[1].split("/")[1]
            egrp = words[-1]
            if not egrp in egrpElmDict.keys():
                egrpElmDict[egrp] = []
            egrpElmDict[egrp].append(elmType)
    egrps = list(egrpElmDict.keys())
    if len(egrps) == 0:
        return []
    egrps.sort()
    addLines = []
    addLines += ["#===================================================================================================\n"]
    addLines += ["#  PROPERTY define\n"]
    propNo = 1
    #solidを確認
    for egrp in egrps:
        elmNames = egrpElmDict[egrp]
        for elm in elmNames:
            #solid部を作成
            if elmTypeDict[elm] == "solid":
                addLines += ["#---1----|----2----|----3----|----4----|----5----|----6----|----7----|----8----|----9----|---10----|\n"]
                addLines += ["#  PROP-" + str(propNo) + "  ElementGroup " + egrp + "\n"]
                addLines += ["#--------.---------.---------.---------.\n"]
                addLines += ["/PROP/SOLID/" + str(propNo) + "\n"]
                addLines += ["solid prperty  ElementGroup " + egrp + "\n"]
                addLines += ["#   Isolid                                Itetra10             Itetra4\n"]
                if elm == "BRICK":
                    addLines += ["         1\n"]
                elif elm == "TETRA4":
                    addLines += ["                                                                  1000\n"]
                radConts.append("/PROP/SOLID/" + str(propNo) + "  " + elm + "  ElementGroup " + egrp + "\n")
                propNo += 1
            #shellを作成
            elif elmTypeDict[elm] == "shell":
                th, N = shellThDict[egrp]
                [th] = checkStrLength10([th])
                addLines += ["#---1----|----2----|----3----|----4----|----5----|----6----|----7----|----8----|----9----|---10----|\n"]
                addLines += ["#  PROP-" + str(propNo) + "  ElementGroup " + egrp + "\n"]
                addLines += ["#--------.---------.---------.---------.\n"]
                addLines += ["/PROP/SHELL/" + str(propNo) + "\n"]
                addLines += ["shell prperty  ElementGroup " + egrp + "\n"]
                elmNames = egrpElmDict[egrp]
                if elm == "SHELL":
                    addLines += ["#   Ishell    Ismstr     Ish3n    Idrill                            P_Thick_Fail\n"]
                    addLines += ["         2\n"]
                    addLines += ["#                 Hm                  Hf                  Hr                  Dm                  Dn\n"]
                    addLines += ["\n"]
                    addLines += ["#        N   Istrain               Thick              Ashear              Ithick     Iplas\n"]
                    addLines += [("%10s" % N) + (" " * 20) + ("%10s" % th) + "\n"]
                elif elm == "SH3N":
                    addLines += ["#   Ishell    Ismstr     Ish3n    Idrill                            P_Thick_Fail\n"]
                    addLines += ["                             2\n"]
                    addLines += ["#                 Hm                  Hf                  Hr                  Dm                  Dn\n"]
                    addLines += ["\n"]
                    addLines += ["#        N   Istrain               Thick              Ashear              Ithick     Iplas\n"]
                    addLines += [("%10s" % N) + (" " * 20) + ("%10s" % th) + "\n"]
                radConts.append("/PROP/SHELL/" + str(propNo) + "  " + elm + "  ElementGroup " + egrp + "\n")
                propNo += 1
            elif elmTypeDict[elm] == "sph":
                massP, elmSize = getPointMassElementSizeOfSph(egrp, fistrDicts)
                massP, elmSize = checkStrLength20([str(massP), str(elmSize)])
                addLines += ["#---1----|----2----|----3----|----4----|----5----|----6----|----7----|----8----|----9----|---10----|\n"]
                addLines += ["#  PROP-" + str(propNo) + "  ElementGroup " + egrp + "\n"]
                addLines += ["#--------.---------.---------.---------.\n"]
                addLines += ["#  mp: set calculated element mass from cell volume and density\n"]
                addLines += ["#   h: set element size from FrontISTR\n"]
                addLines += ["/PROP/TYPE34/" + str(propNo) + "\n"]
                addLines += ["SPH property  ElementGroup " + egrp + "\n"]
                addLines += ["#                 mp                  qa                  qb            alpha_cs   skew_ID      h_ID\n"]
                addLines += [("%20s" % massP) + "                   2                   1                   0         0         0\n"]
                addLines += ["#    order                   h             xi_stab\n"]
                addLines += ["         0" + ("%20s" % elmSize) + "                   0\n"]
                radConts.append("/PROP/TYPE34/" + str(propNo) + "  " + elm + "  ElementGroup " + egrp + "\n")
                propNo += 1
    return addLines

#
#  getPointMassElementSizeOfSph
def getPointMassElementSizeOfSph(egrp, fistrDicts):
    """ egrpの要素の質量と要素サイズを取得して返す"""
    global meshConts
    [nodeDict, elmDict, ngrpDict, surfDict, egrpDict, 
     matDict, shellThDict, contactDict, nodeElmDict] = fistrDicts
    #最初の要素を取得
    elmNo = list(egrpDict[egrp])[0]
    #node取得
    nodeNos, elmType = elmDict[elmNo]
    #要素typeを取得
    elmType = str(elmType)
    eType = cm.elmContsDict[elmType][0]
    #egrpの体積を算出
    V = 0.0
    for elmNo in egrpDict[egrp]:
        nodeNos = elmDict[elmNo][0]
        elmName = elmDict[elmNo][1]
        elmConts = cm.elmContsDict[str(elmName)]
        locs = list(map(lambda x: nodeDict[x], nodeNos))
        if elmConts[2] == "tri":
            a = geom.triA(locs)
            t = shellThDict[egrp][0]
            v = float(a) * float(t)
        elif elmConts[2] == "quad":
            a, g = geom.polygonAG(locs)
            t = shellThDict[egrp][0]
            v = float(a) * float(t)
        elif elmConts[2] == "tet":
            v = geom.tetV(locs)
        elif elmConts[2] == "hex":
            v = geom.hexV(locs)
        V += v
    #粒子数を取得
    for line in meshConts:
        words = line.split()
        if words[1][:len("/SPHCEL/")] == "/SPHCEL/":
            if words[-1] == egrp:
                nSph = int(words[3])
                break
    #粒子1個の質量を取得
    rho = float(matDict[egrp][3])
    massP = V * rho / nSph
    #要素sizeを取得
    size = geom.length(locs[0], locs[1])
    return massP, size

#
#  createPartLines
#------------------
def createPartLines():
    """ PARTを作成"""
    global meshConts, radConts
    #partIDを取得
    egrpPartNoElms = []
    elmNames = elmTypeDict.keys()
    for line in meshConts:
        words = line.split()
        ws = words[1].split("/")
        elmName = ws[1]
        if elmName in elmNames:
            partNo = ws[2]
            egrp = words[-1]
            egrpPartNoElms.append([egrp, partNo, elmName])
    if len(egrpPartNoElms) == 0:
        return []
    egrpPartNoElms.sort()
    addLines = []
    addLines += ["#===================================================================================================\n"]
    addLines += ["#  PART define\n"]
    for egrp, partNo, elmName in egrpPartNoElms:
        propNo = getPropertyNoFromRadConts(egrp, elmName)
        matNo = getMaterialNoFromRadConts(egrp)
        addLines += ["#---1----|----2----|----3----|----4----|----5----|----6----|----7----|----8----|----9----|---10----|\n"]
        addLines += ["#  PART-" + str(partNo) + "  " + elmName + "  ElementGroup " + egrp + "\n"]
        addLines += ["#--------.---------.---------.---------.\n"]
        addLines += ["/PART/" + partNo + "\n"]
        addLines += [elmName + "\n"]
        addLines += ["#  prop_ID    mat_ID\n"]
        addLines += [("%10s" % propNo) + ("%10s" % matNo) + "\n"]
        radConts.append("/PART/" + partNo + "  " + elmName + "  ElementGroup " + egrp + "\n")
    return addLines

#
#  getPropertyNoFromRadConts
def getPropertyNoFromRadConts(egrp, elmName):
    """ radContsからegrp、elmNameが同じpropIdを取得する。"""
    global radConts
    propNo = ""
    for line in radConts:
        words = line.split()
        ws = words[0].split("/")
        if len(ws) > 1:
            if ws[1] == "PROP":
                egrpProp = words[-1]
                elmProp = words[1]
                if egrpProp == egrp and elmProp == elmName:
                    propNo = ws[-1]
                    break
    return propNo

#
#  getMaterialNoFromRadConts
def getMaterialNoFromRadConts(egrp):
    """ radContsからegrpのmatNoを取得する"""
    global radConts
    matNo = ""
    for line in radConts:
        words = line.split()
        ws = words[0].split("/")
        if len(ws) > 1:
            if ws[1] == "MAT":
                egrpProp = words[-1]
                if egrpProp == egrp:
                    try:
                        a = int(ws[-2])
                        matNo = ws[-2]
                    except:
                        a = int(ws[-1])
                        matNo = ws[-1]
                    break
    return matNo

#
#  createFixBoundaryLines
#-------------------------
def createFixBoundaryLines(fixNameDict):
    """ BOUNDARY(0,0,0)の行を作成"""
    global radConts
    names = list(fixNameDict.keys())
    if len(names) == 0:
        return []
    names.sort()
    addLines = []
    addLines += ["#===================================================================================================\n"]
    addLines += ["#  BOUNDARY conditions\n"]
    bcNo = 1
    for name in names:
        grpID = fixNameDict[name][1]
        addLines += ["#---1----|----2----|----3----|----4----|----5----|----6----|----7----|----8----|----9----|---10----|\n"]
        addLines += ["#  BC-" + grpID + "  grpName " + name + "\n"]
        addLines += ["#--------.---------.---------.---------.\n"]
        addLines += ["/BCS/" + str(bcNo) + "\n"]
        addLines += ["nodeGroup " + name + "\n"]
        addLines += ["#  Tra rot   skew_ID  grnod_ID\n"]
        fix = fixNameDict[name][0]
        addLines += ["   " + fix + " 000" + ("%10s" % "0") + ("%10s" % grpID) + "\n"]
        radConts.append("/BCS/" + str(bcNo) + "  nodeGroup " + name + "\n")
        bcNo += 1
    return addLines

#
#  createInitialVelocityLines
#-----------------------------
def createInitialVelocityLines(iniVelDict):
    """ 初速を設定"""
    global radConts
    names = list(iniVelDict.keys())
    if len(names) == 0:
        return []
    names.sort()
    addLines = []
    addLines += ["#===================================================================================================\n"]
    addLines += ["#  INITIAL VELOCITY\n"]
    velNo = 1
    for name in names:
        grpID = iniVelDict[name][1]
        vx, vy, vz = iniVelDict[name][0]
        vx, vy, vz = checkStrLength20([vx, vy, vz])
        addLines += ["#---1----|----2----|----3----|----4----|----5----|----6----|----7----|----8----|----9----|---10----|\n"]
        addLines += ["#  INIVEL-" + grpID + "  grpName " + name + "\n"]
        addLines += ["#--------.---------.---------.---------.\n"]
        addLines += ["/INIVEL/TRA/" + str(velNo) + "\n"]
        addLines += ["nodeGroup " + name + "\n"]
        addLines += ["#                 Vx                  Vy                  Vz   grnd_ID   skew_id\n"]
        addLines += [("%20s" % vx) + ("%20s" % vy) + ("%20s" % vz) + ("%10s" % grpID) + "\n"]
        radConts.append("/INIVEL/TRA/" + str(velNo) + "  nodeGroup " + name + "\n")
        velNo += 1
    return addLines

#
#  createContactLines
#---------------------
def createContactLines(cntHeaderData, fistrDicts):
    """ contactの行を作成する。"""
    global meshConts, radConts
    [nodeDict, elmDict, ngrpDict, surfDict, egrpDict, 
     matDict, shellThDict, contactDict, nodeElmDict] = fistrDicts
    contacts = list(contactDict.keys())
    if len(contacts) == 0:
        return []
    #contact_algoを取得
    addLines = []
    addLines += ["#===================================================================================================\n"]
    addLines += ["#  CONTACT define\n"]
    cardData = getHeaderDataOfCard(cntHeaderData, "!CONTACT_ALGO")
    header = cardData[0][0]
    algo = pyFistr.getValName(header, "TYPE")
    contacts.sort()
    interNo = 1
    for contact in contacts:
        #該当するcontactを取得
        cardData = getHeaderDataOfCard(cntHeaderData, "!CONTACT")
        for header, data in cardData:
            words = pyFistr.deleteSp(data[0]).split(",")
            cpName = words[0]
            if contact == cpName:
                break
        #摩擦係数fcoeffを取得
        fcoeff = words[1]
        #grnodID, surIDを取得
        names = contactDict[cpName]
        ngID, sgID = getPairIDsFromMeshConts(names)
        #gapMaxを取得（要素サイズの5倍に設定）
        elmSize = getElementSizeOfNgrp(names[0], fistrDicts)
        gapMax = str(elmSize * 5.0)
        (fcoeff, gapMax) = checkStrLength20([fcoeff, gapMax])
        addLines += ["#---1----|----2----|----3----|----4----|----5----|----6----|----7----|----8----|----9----|---10----|\n"]
        addLines += ["#  INTER-" + str(interNo) + "  " + cpName + ": " + ",".join(names) + "\n"]
        addLines += ["#--------.---------.---------.---------.\n"]
        addLines += ["#     Fric: set from FrontISTR\n"]
        addLines += ["#  Gap_max: set from element size of nodeGroup\n"]
        addLines += ["/INTER/TYPE7/" + str(interNo) + "\n"]
        addLines += ["contact between " + names[0] + " and " + names[1] + "\n"]
        addLines += ["#  Slav_id   Mast_id      Istf      Ithe      Igap                Ibag      Idel     Icurv      Iadm\n"]
        addLines += [("%10s" % ngID) + ("%10s" % sgID) + "         0         0         3                   0         1         0         0\n"]
        addLines += ["#          Fscalegap             Gap_max             Fpenmax                         Itied\n"]
        addLines += [("%20d" % 1) + ("%20s" % gapMax) + "                 0.0                             0\n"]
        addLines += ["#              Stmin               Stmax          %mesh_size               dtmin  Irem_gap   Irem_i2\n"]
        addLines += ["                 0.0                 0.0                 0.0                 0.0         0         0\n"]
        addLines += ["#              Stfac                Fric              Gapmin              Tstart               Tstop\n"]
        addLines += ["                 1.0" + ("%20s" % fcoeff) + "                 0.0                 0.0                 0.0\n"]
        addLines += ["#      IBC                        Inacti                VisS                VisF              Bumult\n"]
        addLines += ["       000                             5                 0.0                 0.0                 0.0\n"]
        addLines += ["#    Ifric    Ifiltr               Xfreq     Iform   sens_ID   fct_IDf             AscaleF   fric_ID\n"]
        addLines += ["         0         0                 0.0         0         0         0                 0.0         0\n"]
        radConts.append("/INTER/TYPE7/" + str(interNo) + "  contact " + cpName + " " + ",".join(names) + "\n")
        interNo += 1
    return addLines

#
#  getPairIDsFromMeshConts
def getPairIDsFromMeshConts(pairNames):
    """ 接触pairNameのIDを取得する。"""
    ngrp, sgrp = pairNames
    ngID = getGroupIdFromMeshConts(ngrp, "GRNOD")
    sgID = getGroupIdFromMeshConts(sgrp, "SURF")
    return ngID, sgID

#
#  getElementSizeOfNgrp
def getElementSizeOfNgrp(ngrp, fistrDicts):
    """ ngrpが属する要素のサイズを取得する"""
    [nodeDict, elmDict, ngrpDict, surfDict, egrpDict, 
     matDict, shellThDict, contactDict, nodeElmDict] = fistrDicts
    #最初のnodeNoを取得
    nodeNo = ngrpDict[ngrp][0]
    elmNo = list(nodeElmDict[nodeNo])[0]
    nodeNos = elmDict[elmNo][0]
    #node0, node1の距離を取得
    loc0 = nodeDict[nodeNos[0]]
    loc1 = nodeDict[nodeNos[1]]
    size = geom.length(loc0, loc1)
    return size

#
#  createRadFooter
#------------------
def createRadFooter():
    """ radFileのfooterを作成"""
    addLines  = ["#===================================================================================================\n"]
    addLines += ["/END\n"]
    addLines += ["#--------.---------.---------.---------.---------.---------.---------.---------.---------.---------.\n"]
    return addLines

#--------------------------
#  createRadiossEngineFile
#--------------------------
def createRadiossEngineFile(fistrFiles, radiossFiles, fistrDicts):
    """ RadiossのEngineファイルを作成する"""
    global engConts
    engConts = []
    [meshFile, cntFile, sphEgrp, fistrUnit] = fistrFiles
    [incFile, radFile, yearVersion, unit] = radiossFiles
    [nodeDict, elmDict, ngrpDict, surfDict, egrpDict, 
     matDict, shellThDict, contactDict, nodeElmDict] = fistrDicts
    #cntFile読み込み
    f = open(cntFile, encoding="utf-8"); lines = f.readlines(); f.close()
    cntHeaderData = pyFistr.getHeaderData(lines)
    #EngineHeader作成
    lines  = createEngineHeaderLines(cntFile, meshFile)
    lines += craeteVersionLines(yearVersion)
    lines += createMonRunTfileLines(radFile, cntHeaderData)
    lines += createDTLines(cntHeaderData)
    lines += createAnimData()
    #fileに出力
    engFile = "_".join(radFile.split("_")[:-1]) + "_0001.rad"
    f = open(engFile, "w", encoding="utf-8"); f.writelines(lines); f.close()

#
#  createEngineHeaderLines
#---------------------
def createEngineHeaderLines(cntFile, meshFile):
    """ EngineFileのheaderを作成する"""
    now = datetime.datetime.now()
    dateList = [str(now.year), str(now.month), str(now.day)]
    addLines = []
    addLines  = ["#RADIOSS ENGINE\n"]
    addLines += ["#===================================================================================================\n"]
    addLines += ["#  generated by fistr2rad.py\n"]
    addLines += ["#           from " + meshFile + " and " + cntFile + "\n"]
    addLines += ["#                                                " + "/".join(dateList) + "\n"]
    addLines += ["#--------.---------.---------.---------.---------.---------.---------.---------.---------.---------.\n"]
    return addLines

#
#  craeteVersionLines
#---------------------
def craeteVersionLines(yearVersion):
    """ version行を作成"""
    global engConts
    addLines = []
    addLines += ["#  VERS\n"]
    addLines += ["/VERS/" + yearVersion + "\n"]
    engConts.append("/VERS/" + yearVersion + "\n")
    return addLines

#
#  createMonRunTfileLines
#-----------------
def createMonRunTfileLines(radFile, cntHeaderData):
    """ RUN行を作成"""
    global engConts
    runName = "_".join(radFile.split("_")[:-1])
    endTime = getEndTimeFromCntFile(cntHeaderData)
    [endTime] = checkStrLength10([str(endTime)])
    addLines = []
    addLines += ["#  RUN\n"]
    addLines += ["/RUN/" + runName + "/1\n"]
    addLines += [("%10s" % endTime) + "\n"]
    addLines += ["/MON/ON\n"]
    addLines += ["/TFILE/0\n"]
    engConts.append("/RUN/" + runName + "/1  " + str(endTime) + "\n")
    engConts.append("/MON/ON\n")
    engConts.append("/TFILE/0\n")
    return addLines

#
#  getEndTimeFromCntFile
def getEndTimeFromCntFile(cntHeaderData):
    """ endTimeを取得する"""
    endTime = ""
    for header, data in cntHeaderData:
        words = pyFistr.deleteSp(header).split(",")
        if words[0] == "!DYNAMIC":
            line = data[1]
            ws = pyFistr.deleteSp(line).split(",")
            endStep = int(ws[2])
            DT = float(ws[3])
            endTime = endStep * DT
            break
    return endTime

#
#  createDTLines
#----------------
def createDTLines(cntHeaderData):
    """ 時間ステップの設定"""
    global engConts
    DT, animDT = getDeltaTFromCnt(cntHeaderData)
    DT, animDT = checkStrLength10([str(DT), str(animDT)])
    addLines = []
    addLines += ["#  DT\n"]
    addLines += ["/DT/NODA/CST/0\n"]
    addLines += [("%10s" % "0.9") + ("%10s" % DT) + "\n"]
    addLines += ["/ANIM/DT\n"]
    addLines += [("%10s" % "0.0") + ("%10s" % animDT) + "\n"]
    engConts.append("/DT/NODA/CST/0  " + str(DT) + "\n")
    engConts.append("/ANIM/DT  " + str(animDT) + "\n")
    return addLines

#
#  getDeltaTFromCnt
def getDeltaTFromCnt(cntHeaderData):
    """ deltaT（時間増分）をcntFileから取得"""
    DT = 0
    animDT = 1
    endStep = 1
    for header, data in cntHeaderData:
        words = pyFistr.deleteSp(header).split(",")
        if words[0] == "!DYNAMIC":
            line = data[1]
            ws = pyFistr.deleteSp(line).split(",")
            endStep = int(ws[2])
            DT = float(ws[3])
        elif words[0] == "!WRITE":
            freq = pyFistr.getValName(header, "FREQUENCY")
            if freq == "":
                n = 1
            else:
                n = int(freq)
        if DT > 0 and animDT > 0:
            break
    animDT = DT * n
    if DT == 0:
        #未設定の場合
        return DT, animDT
    nAnim = endStep * DT / animDT
    if nAnim > 1000:
        print()
        print("ERROR: number of animations are too many!!")
        print("  over 1000 animations.")
        print("  check write frequency of FrontISTR.")
        exit()
    return DT, animDT
   
#
#  createAnimData
def createAnimData():
    """ animationに関するデータ行を作成"""
    global engConts
    #要素確認flagを取得
    (shell, solid, sph) = getElementFlag()
    #節点変位
    addLines = []
    addLines += ["/ANIM/VECT/DISP\n"]
    addLines += ["/ANIM/VECT/VEL\n"]
    addLines += ["/ANIM/ELEM/EPSP\n"]
    addLines += ["/ANIM/ELEM/VONM\n"]
    engConts.append("/ANIM/VECT/DISP\n")
    engConts.append("/ANIM/VECT/VEL\n")
    engConts.append("/ANIM/ELEM/EPSP\n")
    engConts.append("/ANIM/ELEL/VONM\n")
    if solid == True:
        addLines += ["/ANIM/BRIC/TENS/STRESS\n"]
        addLines += ["/ANIM/BRIC/TENS/STRAIN\n"]
        #addLines += ["/ANIM/BRIC/TENS/EPSP\n"]
        #addLines += ["/ANIM/BRIC/VONM\n"]
        engConts.append("/ANIM/BRIC/TENS/STRESS\n")
        engConts.append("/ANIM/BRIC/TENS/STRAIN\n")
        #engConts.append("/ANIM/BRIC/TENS/EPSP\n")
        #engConts.append("/ANIM/BRIC/VONM\n")
    if shell == True:
        addLines += ["/ANIM/SHELL/TENS/UPPER\n"]
        addLines += ["/ANIM/SHELL/TENS/LOWER\n"]
        #addLines += ["/ANIM/SHELL/VONM\n"]
        engConts.append("/ANIM/SHELL/TENS/UPPER\n")
        engConts.append("/ANIM/SHELL/TENS/LOWER\n")
        #engConts.append("/ANIM/SHELL/VONM\n")
    return addLines

#
#  getElementFlag
def getElementFlag():
    """ 使用している要素をチェック"""
    global meshConts
    shell = False; solid = False; sph = False
    for line in meshConts:
        words = line.split()
        if words[2] == "Elements":
            ws = words[1].split("/")
            elm = ws[1]
            elmType = elmTypeDict[elm]
            if elmType == "shell":
                shell = True
            elif elmType == "solid":
                solid = True
            elif elmType == "sph":
                sph = True
    return shell, solid, sph

#
#  printConvertResult
#--------------------
def printConvertResult(meshFile, cntFile, incFile, radFile):
    """ mesh変換結果を出力する"""
    global meshConts, radConts, engConts
    msg = ["\n"]
    msg += ["Fistr mesh has been converted to Radioss format.\n"]
    msg += ["\n"]
    msg += [" mesh contents\n"]
    msg += list(map(lambda x: x[1:], meshConts))
    msg += ["\n"]
    msg += [" STRATER contents\n"]
    msg += list(map(lambda x: "   " + x, radConts))
    msg += ["\n"]
    msg += [" ENGINE contents\n"]
    msg += list(map(lambda x: "   " + x, engConts))
    msg += ["\n"]
    msg += ["input files\n"]
    msg += ["  msh: " + meshFile + "\n"]
    msg += ["  cnt: " + cntFile + "\n"]
    msg += ["output files\n"]
    msg += ["  inc: " + incFile + "\n"]
    msg += ["  rad: " + radFile + "\n"]
    engFile = "_".join(radFile.split("_")[:-1]) + "_0001.rad"
    msg += ["       " + engFile + "\n"]
    msg = "".join(msg)
    print(msg)


#--------------------
#  convertFistr2rad
#--------------------
def convertFistr2rad(fistrFiles, radiossFiles):
    """ fistr.mshをrad形式のメッシュに変換する"""
    global meshConts
    [meshFile, cntFile, sphEgrp, fistrUnit] = fistrFiles
    [incFile, radFile, yearVersion, unit] = radiossFiles
    #fistrのメッシュHeaderDataを取得
    f = open(meshFile, encoding="utf-8"); lines = f.readlines(); f.close()
    headerData = pyFistr.getHeaderNumData(lines)
    f = open(cntFile, encoding="utf-8"); lines = f.readlines(); f.close()
    cntHeaderData = pyFistr.getHeaderData(lines)

    #node, element, ngrp, egrpを取得
    nodeDict = getNodeDataDict(headerData)
    elmDict = getElementDataDict(headerData)
    nodeElmDict = getNodeElementDict(elmDict)
    ngrpDict = getNodeGroupDict(headerData)
    surfDict = getSurfaceGroupDict(headerData, elmDict)
    egrpDict = getElementGroupDict(headerData)
    nodeDict, elmDict, ngrpDict = deleteDummyNodesChangeElmType(nodeDict, elmDict, ngrpDict)
    shellThDict = getShellThicknessDict(headerData)    
    contactDict = getContactPair(headerData)
    #nodeDict, elmDict, ngrpDict = renumberNode(nodeDict, elmDict, ngrpDict)
    #rad形式に要素を変換
    elmRadDict = convertElement(elmDict)
    #egrpと材料定数を取得
    matDict = getEgrpMaterialDictFromCnt(cntHeaderData, headerData)
    fistrDicts = [nodeDict, elmDict, ngrpDict, surfDict, egrpDict,
                  matDict, shellThDict, contactDict, nodeElmDict]

    #incFileの内容を作成
    createRadiossIncMeshFile(fistrDicts, elmRadDict, sphEgrp, incFile)
    #starterFileを作成
    createRadiossStarterFile(fistrFiles, radiossFiles, fistrDicts)
    #engineFileを作成
    createRadiossEngineFile(fistrFiles, radiossFiles, fistrDicts)

    #結果を標準出力に出力
    printConvertResult(meshFile, cntFile, incFile, radFile)

#
#  checkFileNames
#----------------
def checkFileNames(fistrFiles, radiossFiles):
    [meshFile, cntFile, sphEgrp] = fistrFiles
    [incFile, radFile, yearVersion, unitNo] = radiossFiles
    """ file名をチェック"""
    if meshFile.split(".")[-1] != "msh":
        meshFile += ".msh"
    if cntFile.split(".")[-1] != "cnt":
        cntFile += ".cnt"
    if incFile.split(".")[-1] != "inc":
        incFile += ".inc"
    if radFile.split("_")[-1] != "0000.rad":
        radFile += "_0000.rad"
    fistrUnit = getCurrDirUnit()
    if unitNo == "":
        unit = fistrUnit[:]
    else:
        unit = unitDict[unitNo]
        fistrUnit = unit
    fistrFiles = [meshFile, cntFile, sphEgrp, fistrUnit]
    radiossFiles = [incFile, radFile, yearVersion, unit]
    return fistrFiles, radiossFiles

#
#  getCurrDirUnit
def getCurrDirUnit():
    """ currDir内の単位を取得する"""
    unit = "kg_m_s"
    fileName = "easyIstrSub_data"
    if os.path.exists(fileName) == True:
        f = open(fileName, encoding="utf-8"); lines = f.readlines(); f.close()
        for line in lines:
            words = line.split()
            if len(words) > 1:
                if words[0] == "dimension":
                    unit = words[1]
                    break
    fistrUnit = unit.split("_")
    return fistrUnit

#
#  printHelp
#-------------
def printHelp():
    cont = """
------------ fistr2rad.py のヘルプ --------------------------------------
FrontISTR形式のmsh, cntファイルから、
Radioss形式のSTARTER, ENGINEファイルに変換する。

<使い方>
fistr2rad.py [option]
option
    -w <workFolder> :対象のworkFolderを指定する。
                     省略可:省略時はcurrDir。
    -i <header>     :msh, cntファイルのheader名を設定。
                     省略可:省略時はcurrDir内の「hecmw_ctrl.dat」から取得
    -msh <mshFile>  :FistrのmshFile名
                     省略可:省略時はcurrDir内の「hecmw_ctrl.dat」から取得
    -cnt <cntFile>  :FistrのcntFile名
                     省略可:省略時はcurrDir内の「hecmw_ctrl.dat」から取得
    -sph <EGRP>     :指定した要素group名が構成する節点座標を
                     使って、粒子SPHを作成する。
    -o <header>     :inc, radファイルのheader名を設定。
                     省略可:省略時はcntFileのheader名を設定
    -inc <incFile>  :Radioss形式のincludeするfile名
                     省力可:省略時は「<cntFileのheader>_0000.inc」を設定
    -rad <radFile>  :Radioss形式のheader名
                     省略可:省略時は「<cntFileのheader>_0000.rad」を設定
    -v <version>    :年のversionを指定。
                     省略可:省略時は、2022
    -u <unitNo>     :使用する単位系の番号を指定
                     1:MKS, 2:ton,mm,s, 3:kg,mm,ms, 4:CGS
                     省略可:省略時はcurrDir内の設定unitを読み取る。
                     読み取れない場合は、MKS。
    -h, --help      :ヘルプを出力

<使用例>
fistr2rad.py
fistr2rad.py -msh FistrModel.msh -inc FistrModel.inc
"""
    print(cont)

#
#  getOption
#-------------
def getOption():
    """ optionを取得"""
    workFolder = ""
    cntFile, meshFile = "", ""
    incFile = ""
    radFile = ""
    sphEgrp = ""
    yearVersion = "2022"
    unit = ""
    help = "off"
    i = 1
    while i < len(sys.argv):
        if sys.argv[i] == "-w":
            i += 1
            workFolder = sys.argv[i]
        elif sys.argv[i] == "-i":
            i += 1
            meshFile = sys.argv[i] + ".msh"
            cntFile = sys.argv[i] + ".cnt"
        elif sys.argv[i] == "-msh":
            i += 1
            meshFile = sys.argv[i]
        elif sys.argv[i] == "-cnt":
            i += 1
            cntFile = sys.argv[i]
        elif sys.argv[i] == "-o":
            i += 1
            incFile = sys.argv[i] + "_0000.inc"
            radFile = sys.argv[i] + "_0000.rad"
        elif sys.argv[i] == "-inc":
            i += 1
            incFile = sys.argv[i]
        elif sys.argv[i] == "-rad":
            i += 1
            radFile = sys.argv[i]
        elif sys.argv[i] == "-sph":
            i += 1
            sphEgrp = sys.argv[i]
        elif sys.argv[i] == "-v":
            i += 1
            yearVersion = sys.argv[i]
        elif sys.argv[i] == "-u":
            i += 1
            unit = sys.argv[i]
        elif sys.argv[i] == "-h" or sys.argv[i] == "--help":
            printHelp()
            help = "on"
            break
        else:
            printHelp()
            help = "on"
        i += 1
    #currDir移動
    if workFolder != "":
        os.chdir(workFolder)
    #meshFile、cntFile名を取得
    if meshFile == "" or cntFile == "":
        cFile, mFile = getCntMshFileName(os.getcwd())
    if meshFile == "":
        meshFile = mFile
    if cntFile == "":
        cntFile = cFile
    #inc, radのファイル名を取得
    if incFile == "" or radFile == "":
        header = ".".join(cntFile.split(".")[:-1])
    if incFile == "":
        incFile = header + "_0000.inc"
    if radFile == "":
        radFile = header + "_0000.rad"
    fistrFiles = [meshFile, cntFile, sphEgrp]
    radiossFiles = [incFile, radFile, yearVersion, unit]
    return fistrFiles, radiossFiles, help


if __name__ == "__main__":
    import gettext
    gettext.install("easyistr", os.getenv("LOCALE_DIR"))
    #_ = gettext.gettext

    fistrFiles, radiossFiles, help = getOption()
    if help == "on":
        exit()
    fistrFiles, radiossFiles = checkFileNames(fistrFiles, radiossFiles)
    convertFistr2rad(fistrFiles, radiossFiles)
